/*
***********************************************************************
* COPYRIGHT AND WARRANTY INFORMATION
*
* Copyright 2001, International Telecommunications Union, Geneva
*
* DISCLAIMER OF WARRANTY
*
* These software programs are available to the user without any
* license fee or royalty on an "as is" basis. The ITU disclaims
* any and all warranties, whether express, implied, or
* statutory, including any implied warranties of merchantability
* or of fitness for a particular purpose.  In no event shall the
* contributor or the ITU be liable for any incidental, punitive, or
* consequential damages of any kind whatsoever arising from the
* use of these programs.
*
* This disclaimer of warranty extends to the user of these programs
* and user's customers, employees, agents, transferees, successors,
* and assigns.
*
* The ITU does not represent or warrant that the programs furnished
* hereunder are free of infringement of any third-party patents.
* Commercial implementations of ITU-T Recommendations, including
* shareware, may be subject to royalty fees to patent holders.
* Information regarding the ITU-T patent policy is available from
* the ITU Web site at http://www.itu.int.
*
* THIS IS NOT A GRANT OF PATENT RIGHTS - SEE THE ITU-T PATENT POLICY.
************************************************************************
*/

/*!
 ***********************************************************************
 *  \mainpage
 *     This is the H.26L encoder reference software. For detailed documentation
 *     see the comments in each file.
 *
 *  \author
 *     The main contributors are listed in contributors.h
 *
 *  \version
 *     JM 1.6
 *
 *  \note
 *     tags are used for document system "doxygen"
 *     available at http://www.doxygen.org
 */
/*!
 *  \file
 *     fgs_encoder.c
 *  \brief
 *     TML encoder project main
 *  \author
 *   - Joo Ascenso               <joao.ascenso@lx.it.pt>
 *
 ***********************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <limits.h>
#include <math.h>

#include "global.h"
#include "memalloc.h"
#include "fgs_encoder.h"

//#define FGS_TRACE
#define FGS_TRACE_PATH "d:\\stat.txt"
#ifdef FGS_TRACE
int gtr;
#endif
// Bitstream buffer for FGS coding
Bitstream *pFGSBitstream;
// FGS coding data
FGSStruct *pFGSData;

const int fgs_zigzag[16] = {    //array to indicate the zigzag scan order
  0, 1, 5, 6,
  2, 4, 7,12,
  3, 8,11,13,
  9,10,14,15,
};

static const int fgs_quant_coef[4][4] = {
{16384,10486,16384,10486},
{10486, 6453,10486, 6453},
{16384,10486,16384,10486},
{10486, 6453,10486, 6453},
};


/**************************************************************************
ULVC Tables to encode the cbp index and the (run,eop) codes
***************************************************************************/

const unsigned int table_cbp_yyyy[6][16] = {
	{0x01,0x04,0x05,0x07,0x03, 0x09,0x0A,0x0D,0x02,0x0B,0x08,0x0F,0x06,0x0E,0x0C,0x010},
  {1,5,5,5,3,7,7,7,3,7,7,7,5,7,7,9},

  {0x01,0x03,0x05,0x07,0x04,0x08,0x0B,0x0C,0x02,0x0D,0x09,0x0E,0x06,0x0F,0x010,0x0A},
  {1,3,5,5,5,7,7,7,3,7,7,7,5,7,9,7},

  {0x01,0x02,0x03,0x06,0x04,0x08,0x0B,0x0F,0x05,0x0A,0x09,0x0E,0x07,0x0C,0x0D,0x010},
  {1,3,3,5,5,7,7,7,5,7,7,7,5,7,7,9},
};

const unsigned int table_cbp_yyyyuv[6][48] = {
  {0x01,0x03,0x04,0x08,0x05,0x0A,0x010,0x014,0x06,0x0F,0x09,0x013,0x07,0x015,0x012,0x016,0x011,0x022,
    0x024,0x026,0x023,0x027,0x02B,0x030,0x025,0x02C,0x028,0x02A,0x029,0x02E,0x02D,0x02F,0x02,0x0C,0x0B,
    0x019,0x0E,0x01A,0x01B,0x021,0x0D,0x01C,0x017,0x01E,0x018,0x020,0x01D,0x01F},
  {1,3,5,7,5,7,9,9,5,7,7,9,5,9,9,9,9,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,3,7,7,9,7,9,9,11,7,9,
  9,9,9,11,9,9},

  {0x01,0x05,0x07,0x09,0x06,0x0A,0x01A,0x014,0x04,0x018,0x0C,0x015,0x08,0x016,0x011,0x03,0x0B,0x02D,0x027,
    0x02B,0x02A,0x02E,0x030,0x024,0x028,0x02F,0x02C,0x025,0x029,0x026,0x023,0x013,0x0D,0x020,0x01D,0x01C,
    0x01F,0x01B,0x022,0x012,0x1E,0x021,0x019,0x0E,0x017,0x010,0x0F,0x02},
  {1,5,5,7,5,7,9,9,5,9,7,9,7,9,9,3,7,11,11,11,11,11,11,11,11,11,11,11,11,11,11,9,7,11,9,9,9,9,11,9,9,11,
  9,7,9,9,7,3},

  {0x01,0x04,0x07,0x0D,0x06,0x0F,0x01B,0x013,0x05,0x01A,0x010,0x014,0x09,0x015,0x012,0x03,0x020,0x02B,0x02E,
    0x027,0x02C,0x02A,0x02F,0x026,0x02D,0x030,0x029,0x025,0x028,0x023,0x024,0x011,0x08,0x01E,0x01C,0x019,
    0x01F,0x018,0x022,0x0E,0x01D,0x021,0x017,0x0C,0x016,0x0B,0x0A,0x02},
  {1,5,5,7,5,7,9,9,5,9,9,9,7,9,9,3,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,9,7,9,9,9,9,9,11,7,9,11,9,7,
    9,7,7,3},
};

const unsigned int table_dc_lum_uvlc[6][32] = {
  {0x07,0x0E,0x011,0x015,0x016,0x017,0x018,0x019,0x01A,0x01B,0x01C,0x01D,0x01E,0x01F,0x020,0x01,0x02,0x03,
    0x04,0x05,0x08,0x06,0x0A,0x0C,0x0B,0x09,0x0D,0x010,0x0F,0x013,0x012,0x014},
  {5,7,9,9,9,9,9,9,9,9,9,9,9,9,11,1,3,3,5,5,7,5,7,7,7,7,7,9,7,9,9,9},

  {0x02,0x04,0x05,0x08,0x0C,0x0E,0x012,0x014,0x016,0x018,0x01B,0x01D,0x01E,0x01F,0x020,0x01,0x03,0x06,0x07,
    0x09,0x0A,0x0B,0x0D,0x0F,0x011,0x010,0x013,0x015,0x017,0x019,0x01A,0x01C},
  {3,5,5,7,7,7,9,9,9,9,9,9,9,9,11,1,3,5,5,7,7,7,7,7,9,9,9,9,9,9,9,9},

  {0x01,0x02,0x03,0x05,0x07,0x09,0x0C,0x0F,0x011,0x014,0x017,0x019,0x01C,0x01E,0x020,0x0E,0x04,0x06,0x08,
    0x0A,0x0B,0x0D,0x010,0x012,0x013,0x015,0x016,0x018,0x01A,0x01B,0x01D,0x01F},
  {1,3,3,5,5,7,7,7,9,9,9,9,9,9,11,7,5,5,7,7,7,7,9,9,9,9,9,9,9,9,9,9},
};

const unsigned int table_dc_chr_uvlc[6][8] = {
  {0x02,0x06,0x08,0x03,0x01,0x04,0x05,0x07},
  {3,5,7,3,1,5,5,5},

  {0x01,0x05,0x08,0x03,0x02,0x04,0x06,0x07},
  {1,5,7,3,3,5,5,5},

  {0x01,0x04,0x08,0x06,0x02,0x03,0x05,0x07},
  {1,5,7,5,3,3,5,5},
};

const unsigned int table_ac_uvlc[6][32] = {
  {0x06,0x02,0x05,0x0A,0x0C,0x0D,0x012,0x013,0x015,0x017,0x01A,0x01C,0x01D,0x01F,0x020,0x01,0x07,0x04,0x03,
    0x08,0x0B,0x09,0x0E,0x0F,0x010,0x011,0x016,0x014,0x018,0x01B,0x019,0x01E},
  {5,3,5,7,7,7,9,9,9,9,9,9,9,9,11,1,5,5,3,7,7,7,7,7,9,9,9,9,9,9,9,9},

  {0x02,0x01,0x03,0x05,0x09,0x0C,0x0F,0x011,0x014,0x015,0x018,0x01A,0x01D,0x01F,0x020,0x07,0x04,0x06,0x08,
    0x0A,0x0B,0x0D,0x0E,0x010,0x012,0x013,0x016,0x017,0x019,0x01C,0x01B,0x01E},
  {3,1,3,5,7,7,7,9,9,9,9,9,9,9,11,5,5,5,7,7,7,7,7,9,9,9,9,9,9,9,9,9},

  {0x01,0x02,0x03,0x05,0x07,0x0A,0x0C,0x010,0x012,0x015,0x017,0x019,0x01B,0x01E,0x020,0x0E,0x04,0x06,0x08,
    0x09,0x0B,0x0D,0x0F,0x011,0x013,0x014,0x016,0x018,0x01A,0x01D,0x01C,0x01F},
  {1,3,3,5,5,7,7,9,9,9,9,9,9,9,11,7,5,5,7,7,7,7,7,9,9,9,9,9,9,9,9,9},
};

/*!
 ***********************************************************************
 * \brief
 *    initalizes the encoding process... opens files, allocates memory... etc...
 * \param pimg
 *    ImageParameters
 * \param pinp
 *    InputParameters
 * \return
 *    none
 ***********************************************************************
 */
void initFGS (ImageParameters *pimg,InputParameters *pinp)
{
  int a,nblocks;
  //maximum value no compression at all
  //const int buffer_size = ROUND((MAX_BP_LEVEL*img->width*img->height*1.5)/8);
  const int buffer_size = 7000000;
  printf("FGS encoder initialization\n");
  //alloc space for fgs encoding data
  pFGSData = (FGSStruct *) calloc(1, sizeof(FGSStruct));
  //open bitstream file
  if(NULL == (pFGSData->pFGSFile=fopen(pinp->FGSBitstreamFile, "wb")))
  {
    snprintf (errortext, ET_SIZE, "Cannot open fgs bitstream file '%s'",pinp->FGSBitstreamFile);
    error(errortext,500);
  }
  //open stat file
  if(NULL == (pFGSData->pFGSTraceFile=fopen(FGS_TRACE_PATH, "w")))
  {
    snprintf (errortext, ET_SIZE, "Cannot open fgs bitstream file '%s'",pinp->FGSBitstreamFile);
    error(errortext,500);
  }
  //alloc residual images  
  get_mem2Dint(&pFGSData->Ydiff_ac, img->height, img->width);
  get_mem2Dint(&pFGSData->Udiff_ac, img->height_cr, img->width_cr);
  get_mem2Dint(&pFGSData->Vdiff_ac, img->height_cr, img->width_cr);

  get_mem2Dint(&pFGSData->Ydiff_dc, img->height>>2, img->width>>2);
  pFGSData->Udiff_dc = (int *) calloc((img->height_cr>>2)*(img->width_cr>>2), sizeof(int));
  pFGSData->Vdiff_dc = (int *) calloc((img->height_cr>>2)*(img->width_cr>>2), sizeof(int));

  //alloc bitplanes
  get_mem2D(&pFGSData->YBitplane_ac, img->height, img->width);
  get_mem2D(&pFGSData->UBitplane_ac, img->height_cr, img->width_cr);
  get_mem2D(&pFGSData->VBitplane_ac, img->height_cr, img->width_cr);

  get_mem2D(&pFGSData->YBitplane_dc, img->height>>2, img->width>>2);
  pFGSData->UBitplane_dc = (byte *) calloc((img->height_cr>>2)*(img->width_cr>>2), sizeof(byte));
  pFGSData->VBitplane_dc = (byte *) calloc((img->height_cr>>2)*(img->width_cr>>2), sizeof(byte));

  //alloc bitstream buffer
  pFGSBitstream = (Bitstream *) calloc(1, sizeof(Bitstream));
  pFGSBitstream->streamBuffer = (byte *) calloc(buffer_size, sizeof(byte));
  if (pFGSBitstream == NULL || pFGSBitstream->streamBuffer == NULL)
  {
    snprintf(errortext, ET_SIZE, "Memory allocation for FGS bitstream buffer failed");
    error(errortext, 100);
  }
  //init bitstream buffer
  initializeBitstream(pFGSBitstream);

  pFGSData->hDCTMBlocks = pimg->height/MBSIZE;
  pFGSData->wDCTMBlocks = pimg->width/MBSIZE;

  nblocks = pFGSData->hDCTMBlocks*pFGSData->wDCTMBlocks;

  //allocate structure with encoding data for luminance
  pFGSData->block_data_lum_ac = (enh_block **)calloc(4*nblocks,sizeof(enh_block *));
  if(pFGSData->block_data_lum_ac == NULL)
    no_mem_exit("Memory allocation failure: initFGS");

  for(a=0;a < 4*nblocks;a++)
  {
    pFGSData->block_data_lum_ac[a] = (enh_block *)calloc(1,sizeof(enh_block));
    if(pFGSData->block_data_lum_ac[a] == NULL)
      no_mem_exit("Memory allocation failure: initFGS");
  }

  pFGSData->block_data_lum_dc = (enh_block **)calloc(nblocks,sizeof(enh_block *));
  if(pFGSData->block_data_lum_dc == NULL)
    no_mem_exit("Memory allocation failure: initFGS");

  for(a=0;a < nblocks;a++)
  {
    pFGSData->block_data_lum_dc[a] = (enh_block *)calloc(1,sizeof(enh_block));
    if(pFGSData->block_data_lum_dc[a] == NULL)
      no_mem_exit("Memory allocation failure: initFGS");
  }

  pFGSData->hDCTBlocks_cr = pimg->height_cr/BSIZE;
  pFGSData->wDCTBlocks_cr = pimg->width_cr/BSIZE;

  nblocks = pFGSData->hDCTBlocks_cr*pFGSData->wDCTBlocks_cr;

  //allocate structure with encoding data for chrominance components
  pFGSData->block_data_u_ac = (enh_block **)calloc(nblocks,sizeof(enh_block *));
  pFGSData->block_data_v_ac = (enh_block **)calloc(nblocks,sizeof(enh_block *));

  if(pFGSData->block_data_u_ac == NULL || pFGSData->block_data_v_ac == NULL)
    no_mem_exit("Memory allocation failure: initFGS");

  pFGSData->block_data_u_dc = (enh_block **)calloc(nblocks,sizeof(enh_block *));
  pFGSData->block_data_v_dc = (enh_block **)calloc(nblocks,sizeof(enh_block *));

  if(pFGSData->block_data_u_dc == NULL || pFGSData->block_data_v_dc == NULL)
    no_mem_exit("Memory allocation failure: initFGS");

  for(a=0;a < nblocks;a++)
  {
    pFGSData->block_data_u_ac[a] = (enh_block *)calloc(1,sizeof(enh_block));
    pFGSData->block_data_v_ac[a] = (enh_block *)calloc(1,sizeof(enh_block));
    if(pFGSData->block_data_u_ac[a] == NULL || pFGSData->block_data_v_ac[a] == NULL)
      no_mem_exit("Memory allocation failure: initFGS");

    pFGSData->block_data_u_dc[a] = (enh_block *)calloc(1,sizeof(enh_block));
    pFGSData->block_data_v_dc[a] = (enh_block *)calloc(1,sizeof(enh_block));
    if(pFGSData->block_data_u_dc[a] == NULL || pFGSData->block_data_v_dc[a] == NULL)
      no_mem_exit("Memory allocation failure: initFGS");
  }

  //alloc exception code ... total number of macroblocks
  pFGSData->except_code = (byte *) calloc((int)ceil(pFGSData->hDCTMBlocks*pFGSData->wDCTMBlocks), sizeof(byte));
  if(pFGSData->except_code == NULL)
    no_mem_exit("Memory allocation failure: initFGS");
}

/*!
 ***********************************************************************
 * \brief
 *    encodes a frame
 * \param pimg
 *    ImageParameters
 * \param pinp
 *    InputParameters
 * \return
 *    none
 ***********************************************************************
 */

void encodeFGSFrame (ImageParameters *pimg,InputParameters *pinp)
{
  int max_level,max_level_y,max_level_chr,i,nblocks;
 
  fprintf(stdout,"Encoding the fgs enhancement layer\n");
  
  //create residual images: diff + FDCT
  createResidue(pimg);
  //find maximum levels of the Y,U and V component
  findMaxLevel(pimg);

  //find maximum level for all components
  max_level_y = max(pFGSData->maximum_level_y_ac,pFGSData->maximum_level_y_dc);
  max_level_chr = max(pFGSData->maximum_level_chr_ac,pFGSData->maximum_level_chr_dc);
  pFGSData->maximum_level = max(max_level_y,max_level_chr);

  //encode frame header
  outputFGSFrameHeader(pFGSBitstream,pimg->tr,pinp->NumberOfBitplanes);

  //do not encode bp larger than the specified in cfg
  if (pinp->NumberOfBitplanes > pFGSData->maximum_level)
        max_level = pFGSData->maximum_level;
  else
        max_level = pinp->NumberOfBitplanes;

  //clean exception buffer
  nblocks = pFGSData->hDCTMBlocks*pFGSData->wDCTMBlocks;
  memset(pFGSData->except_code,0,nblocks*sizeof(byte));

  //put all cbp's to 0 => meaning: do not code
  for ( i=0; i < (4*nblocks); i++ ) pFGSData->block_data_lum_ac[i]->cbp=0;
  for ( i=0; i < nblocks; i++ ) pFGSData->block_data_lum_dc[i]->cbp=0;

  nblocks = pFGSData->hDCTBlocks_cr*pFGSData->wDCTBlocks_cr;
  for ( i=0; i < nblocks; i++ ) 
  {
    pFGSData->block_data_u_ac[i]->cbp=0;
    pFGSData->block_data_v_ac[i]->cbp=0;
    pFGSData->block_data_u_dc[i]->cbp=0;
    pFGSData->block_data_v_dc[i]->cbp=0;
  }

#ifdef FGS_TRACE
  fprintf(pFGSData->pFGSTraceFile,"Frame number %d\n",pimg->tr);
  fflush(pFGSData->pFGSTraceFile);
#endif

  initEncode();
  for (i=0; i<max_level; i++ )
  {
    //start code
    printf("Level %d\n",i);
    writeIntFGS(FGS_VOP_BP_MARKER_HEADER,FGS_VOP_BP_MARKER_HEADER_LENGTH,pFGSBitstream);
    writeIntFGS(i,FGS_VOP_BP_MARKER_ID_LENGTH,pFGSBitstream);
    calcBitplane(pimg,i);
    encodeBitplane(pFGSBitstream,i,max_level);
  }
  flushBuffer(pFGSBitstream,pFGSData->pFGSFile);
  fprintf(stdout,"Fgs enhancement layer coding finished\n");
}

/*!
 ***********************************************************************
 * \brief
 *    outputs the FGS frame header (a la MPEG-4)
 * \param pBitstream
 *    Bitstream
 * \param tref
 *    temporal reference
 * \param nBp
 *    number of bitplanes
 * \return
 *    none
 ***********************************************************************
 */
void outputFGSFrameHeader (Bitstream *pBitstream,int tref,int nBp)
{
  byte tr;
  //start code
  writeIntFGS(FGS_VOP_START_CODE,FGS_VOP_START_CODE_LENGTH,pBitstream);
  //only FGS (I) frames for now
  writeByteFGS(0x00,2,pBitstream);
  //write temporal reference
  fprintf(stdout,"Time...%d\n", tref);
  tr = tref%256; 
#ifdef FGS_TRACE
  gtr = tr;
#endif
  writeByteFGS(tr,8,pBitstream);
  //marker bit
  writeByteFGS(0x01,1,pBitstream);
  //write fgs_vop_max_level_y,u and v
  printf("DC level:%d AC Level: %d\n",pFGSData->maximum_level_y_dc,
    max(pFGSData->maximum_level_y_ac,max(pFGSData->maximum_level_chr_ac,pFGSData->maximum_level_chr_dc)));
  writeByteFGS(pFGSData->maximum_level_y_dc,5,pBitstream);
  writeByteFGS(pFGSData->maximum_level_y_ac,5,pBitstream);

  writeByteFGS(pFGSData->maximum_level_chr_dc,5,pBitstream);
  writeByteFGS(pFGSData->maximum_level_chr_ac,5,pBitstream);
  
  //marker bit
  writeByteFGS(0x01,1,pBitstream);
  writeByteFGS((byte)(nBp),5,pBitstream);
  //fgs_vop_mc_bit_plane_used.. zero for now => not used
/*writeByteFGS(0x00,5,pBitstream);
  //fgs_vop_selective_enhancement.. zero for now => not used
  writeByteFGS(0x00,1,pBitstream);*/
  //insert stuffing bits
  insertStuffing(pBitstream);
}

/*!
 ***********************************************************************
 * \brief
 *    creates the residual image to be encoded
 * \param pimg
 *    image parameters
 * \return
 *    none
 ***********************************************************************
 */

void createResidue(ImageParameters *pimg)
{
  int a,b;
  
  //calculate residual images
  for(a=0; a<pimg->height ; a++)
  {
    for(b=0; b<pimg->width ; b++)
    {
      pFGSData->Ydiff_ac[a][b]= imgY_org[a][b]-imgY[a][b];
    }
  }
  
  for(a=0; a<pimg->height_cr ; a++)
  {
    for(b=0; b<pimg->width_cr ; b++)
    {
      pFGSData->Udiff_ac[a][b]= imgUV_org[0][a][b]-imgUV[0][a][b];
    }
  }
  
  for(a=0; a<pimg->height_cr ; a++)
  {
    for(b=0; b<pimg->width_cr ; b++)
    {
      pFGSData->Vdiff_ac[a][b]=imgUV_org[1][a][b]-imgUV[1][a][b];
    }
  }

  //the image must be a multiple of 8: padding is not implemented...
  if(pimg->height % DCTSIZE != 0 || pimg->width % DCTSIZE != 0)
  {
    snprintf (errortext, ET_SIZE, "Invalid sequence size\nWidth and Height must be a multiple of 8\n");
    error(errortext,500);
  }

  // calculate the DCT on the Y image

  calcftrans_lum(pFGSData->Ydiff_ac,pFGSData->Ydiff_dc,pFGSData->hDCTMBlocks,pFGSData->wDCTMBlocks);
 
  // calculate the DCT on the U e V image
  calcftrans_chr(pFGSData->Udiff_ac,pFGSData->Udiff_dc,pFGSData->hDCTBlocks_cr,pFGSData->wDCTBlocks_cr);
  calcftrans_chr(pFGSData->Vdiff_ac,pFGSData->Vdiff_dc,pFGSData->hDCTBlocks_cr,pFGSData->wDCTBlocks_cr);
}

/*!
 ***********************************************************************
 * \brief
 *    finds the maximum number of levels necessary to represent an image
 * \param pimg
 *    image parameters
 * \return
 *    none
 ***********************************************************************
 */

void findMaxLevel(ImageParameters *pimg)
{
  int a,b,c,d,maxy_ac,maxu_ac,maxv_ac,maxy_dc,maxu_dc,maxv_dc;
  maxv_ac = maxu_ac = maxy_ac = INT_MIN;
  maxy_dc = maxu_dc = maxv_dc = INT_MIN;

  //scan the images for the maximum abs value
  for(a=0; a<pimg->height ; a++)
  {
    for(b=0; b<pimg->width ; b++)
    {
      c = ABS(pFGSData->Ydiff_ac[a][b]);
      if(c > maxy_ac)
      {
        maxy_ac = c;
      }
    }
  }

  for(a=0; a<(pimg->height>>2) ; a++)
  {
    for(b=0; b<(pimg->width>>2) ; b++)
    {
      c = ABS(pFGSData->Ydiff_dc[a][b]);
      if(c > maxy_dc)
      {
        maxy_dc = c;
      }
    }
  }
  
  for(a=0; a<pimg->height_cr ; a++)
  {
    for(b=0; b<pimg->width_cr ; b++)
    {
      c = ABS(pFGSData->Udiff_ac[a][b]);
      d = ABS(pFGSData->Vdiff_ac[a][b]);
      if(c > maxu_ac)
      {
        maxu_ac = c;
      }
      if(d > maxv_ac)
      {
        maxv_ac = d;
      }
    }
  }

  for(a=0; a<(img->height_cr>>2)*(img->width_cr>>2); a++)
  {
    c = ABS(pFGSData->Udiff_dc[a]);
    d = ABS(pFGSData->Vdiff_dc[a]);
    if(c > maxu_dc)
    {
      maxu_dc = c;
    }
    if(d > maxv_dc)
    {
      maxv_dc = d;
    }  
  }
  
  //calculate the number of bitplanes necessary to represent each image
  pFGSData->maximum_level_y_ac = findLevel(maxy_ac);
  pFGSData->maximum_level_y_dc = findLevel(maxy_dc);

  pFGSData->maximum_level_chr_ac = max(findLevel(maxu_ac),findLevel(maxv_ac)); 
  pFGSData->maximum_level_chr_dc = max(findLevel(maxu_dc),findLevel(maxv_dc));
  
}

/*!
 ***********************************************************************
 * \brief
 *    finds a level of an integer (number of bits necessary to represent it)
 * \param mx
 *    integer
 * \return
 *    number of bits
 ***********************************************************************
 */

int findLevel(int mx)
{
  int i;
  i = MAX_BP_LEVEL-1;
  while ((mx &(1<<i))==0 && i>=0 ) 
  {
    i -= 1;
  }
  i++;
  return i;
}

/*!
 ***********************************************************************
 * \brief
 *    calculates the bitplane
 * \param pimg
 *    image parameters
 * \param level
 *    level to indicate which bitplane must be calculated
 * \return
 *    none
 ***********************************************************************
 */

void calcBitplane(ImageParameters *pimg,int level)
{
  int a,b,n;
  int level_shift = pFGSData->maximum_level-1-level;

  if(level_shift < 0) 
  {
    snprintf (errortext, ET_SIZE, "Invalid bitplane level in calcBitplane\n");
    error(errortext,500);
  }
  //obtain y AC bitplane
  pFGSData->skip_y_ac = TRUE;
  if((pFGSData->maximum_level_y_ac-1) >= level_shift)
  {
    for(a=0; a<pimg->height ; a++)
    {
      for(b=0; b<pimg->width ; b++)
      {
        pFGSData->YBitplane_ac[a][b] = ((ABS(pFGSData->Ydiff_ac[a][b]) >> level_shift) & 0x01);
      }
    }
    pFGSData->skip_y_ac = FALSE;
  }

  //obtain u AC bitplane
  pFGSData->skip_chr_ac = TRUE;
  if((pFGSData->maximum_level_chr_ac-1) >= level_shift)
  {
    for(a=0; a<pimg->height_cr ; a++)
    {
      for(b=0; b<pimg->width_cr ; b++)
      {
        pFGSData->UBitplane_ac[a][b] = ((ABS(pFGSData->Udiff_ac[a][b]) >> level_shift) & 0x01);
        pFGSData->VBitplane_ac[a][b] = ((ABS(pFGSData->Vdiff_ac[a][b]) >> level_shift) & 0x01);
      }
    }
    pFGSData->skip_chr_ac = FALSE;
  }

  //obtain y DC bitplane
  pFGSData->skip_y_dc = TRUE;
  if((pFGSData->maximum_level_y_dc-1) >= level_shift)
  {
    for(a=0; a<(pimg->height>>2) ; a++)
    {
      for(b=0; b<(pimg->width>>2) ; b++)
      {
        pFGSData->YBitplane_dc[a][b] = ((ABS(pFGSData->Ydiff_dc[a][b]) >> level_shift) & 0x01);
      }
    }
    pFGSData->skip_y_dc = FALSE;
  }
  
  n = (img->height_cr>>2)*(img->width_cr>>2); 
  //obtain u DC bitplane
  pFGSData->skip_chr_dc = TRUE;
  if((pFGSData->maximum_level_chr_dc-1) >= level_shift)
  {
    for(a=0; a<n; a++)
    {
      pFGSData->UBitplane_dc[a] = ((ABS(pFGSData->Udiff_dc[a]) >> level_shift) & 0x01);
      pFGSData->VBitplane_dc[a] = ((ABS(pFGSData->Vdiff_dc[a]) >> level_shift) & 0x01);
    }
    pFGSData->skip_chr_dc = FALSE;
  }
  //put the bitplanes with zeros ... since they are going to be coded anyway...
  if(pFGSData->skip_chr_ac == FALSE && pFGSData->skip_chr_dc == TRUE)
  {
    memset(pFGSData->UBitplane_dc,0,n*sizeof(byte));
    memset(pFGSData->VBitplane_dc,0,n*sizeof(byte));
  }
}

/*!
 ***********************************************************************
 * \brief
 *    initializes the encoding process
 *    calculates the data level for a frame and the sign is put in a buffer (zig-zag scan)
 * \return
 *    none
 ***********************************************************************
 */

void initEncode()
{
  int a,b,c,d,i,j,k;
  int temp[6],temp1,temp2;
  enh_block **ldata_ac = pFGSData->block_data_lum_ac;
  enh_block **udata_ac = pFGSData->block_data_u_ac;
  enh_block **vdata_ac = pFGSData->block_data_v_ac;

  enh_block **ldata_dc = pFGSData->block_data_lum_dc;
  enh_block **udata_dc = pFGSData->block_data_u_dc;
  enh_block **vdata_dc = pFGSData->block_data_v_dc;
 
  int mx1,mx2,mx3,mx4,mx5,mx6;
 
  i = 0; k = 0;

  //calculate the data level for ac coeficients

  for(a=0; a<(2*pFGSData->hDCTMBlocks); a+=2)
  {
    for(b=0; b<(2*pFGSData->wDCTMBlocks); b+=2)
    {
      mx1 = mx2 = mx3 = mx4 = mx5 = mx6 = 0;

      for(c=0; c<BSIZE; c++)
      {
        for(d=0; d<BSIZE; d++)
        {
          //for each block (lum,chr) of the MB get the absolute value
          temp[0] = ABS(pFGSData->Ydiff_ac[a*BSIZE+c][b*BSIZE+d]);
          temp[1] = ABS(pFGSData->Ydiff_ac[a*BSIZE+c][(b+1)*BSIZE+d]);
          temp[2] = ABS(pFGSData->Ydiff_ac[(a+1)*BSIZE+c][b*BSIZE+d]);
          temp[3] = ABS(pFGSData->Ydiff_ac[(a+1)*BSIZE+c][(b+1)*BSIZE+d]);
     
          temp[4] = ABS(pFGSData->Udiff_ac[(a/2)*BSIZE+c][(b/2)*BSIZE+d]);
          temp[5] = ABS(pFGSData->Vdiff_ac[(a/2)*BSIZE+c][(b/2)*BSIZE+d]);
          
          //find maximum in each block
          if (temp[0] > mx1) 
          {
            mx1 = temp[0];
          }
          if (temp[1] > mx2)
          {
            mx2 = temp[1];
          }
          if (temp[2] > mx3) 
          {
            mx3 = temp[2];
          }
          if (temp[3] > mx4) 
          {
            mx4 = temp[3];
          }
          if (temp[4] > mx5) 
          {
            mx5 = temp[4];
          }
          if (temp[5] > mx6)
          {
            mx6 = temp[5];
          }
        }
      }
      // and store it in the data_level
      ldata_ac[i]->data_level = findLevel(mx1);
      ldata_ac[i+1]->data_level = findLevel(mx2);
      ldata_ac[i+2]->data_level = findLevel(mx3);
      ldata_ac[i+3]->data_level = findLevel(mx4);
      udata_ac[k]->data_level = findLevel(mx5);
      vdata_ac[k]->data_level = findLevel(mx6);
      k++;
      i+=4;
    }
  }

  //for dc components
  i = 0;
  for(a=0; a<pFGSData->hDCTMBlocks; a++)
  {
    for(b=0; b<pFGSData->wDCTMBlocks; b++)
    {
      mx1 = 0;
      //for luminance
      for(c=0; c<TSIZE; c++)
      {
        for(d=0; d<TSIZE; d++)
        {
          temp1 = ABS(pFGSData->Ydiff_dc[a*TSIZE+c][b*TSIZE+d]);
          if (temp1 > mx1) 
          {
            mx1 = temp1;
          }
        }
      }
      ldata_dc[i]->data_level = findLevel(mx1);

      //for chrominance
      mx1 = mx2 = 0;
      for(c=0; c<TSIZE; c++)
      {
        temp1 = ABS(pFGSData->Udiff_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c]);
        temp2 = ABS(pFGSData->Vdiff_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c]);
        if (temp1 > mx1) 
        {
          mx1 = temp1;
        }
        if (temp2 > mx2) 
        {
          mx2 = temp2;
        }
      }
      udata_dc[i]->data_level = findLevel(mx1);
      vdata_dc[i]->data_level = findLevel(mx2);
      i++;
    }
  }

  //get the sign of each DCT coefficient in zig-zag scan order
  i = 0;j=0;
  for(a=0; a<pFGSData->hDCTMBlocks; a++)
  {
    for(b=0; b<pFGSData->wDCTMBlocks; b++)
    {
      scanSignal(4*a,4*b,ldata_ac[i]->sign,pFGSData->Ydiff_ac);
      scanSignal(4*a,4*b+2,ldata_ac[i+1]->sign,pFGSData->Ydiff_ac);
      scanSignal(4*a+2,4*b,ldata_ac[i+2]->sign,pFGSData->Ydiff_ac);
      scanSignal(4*a+2,4*b+2,ldata_ac[i+3]->sign,pFGSData->Ydiff_ac);

      scanSignal(2*a,2*b,udata_ac[j]->sign,pFGSData->Udiff_ac);
      scanSignal(2*a,2*b,vdata_ac[j]->sign,pFGSData->Vdiff_ac);
      i+=4;
      j++;
    }
  }

  //get the sign of each DC chrominance coefficient
  i = 0;
  for(a=0; a<pFGSData->hDCTMBlocks; a++)
  {
    for(b=0; b<pFGSData->wDCTMBlocks; b++)
    {
      for(c=0; c<TSIZE; c++)
      {
        udata_dc[i]->sign[0][c] = SIGN0(pFGSData->Udiff_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c]);
        vdata_dc[i]->sign[0][c] = SIGN0(pFGSData->Vdiff_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c]);
      }
      for(c=0; c<TSIZE; c++)
      {
        for(d=0; d<TSIZE; d++)
        {
          ldata_dc[i]->sign[0][fgs_zigzag[c*TSIZE+d]] = SIGN0(pFGSData->Ydiff_dc[a*TSIZE+c][b*TSIZE+d]);
        }
      }
      i++;
    }
  }
}

//scanning in zigzag order the bitplane to store the signal
void scanSignal(int hIndex,int wIndex,int sign[][TSIZE2],int **bitplane)
{
  int a,b;

  for(a=0; a<TSIZE; a++)
  {
    for(b=0; b<TSIZE; b++)
    {
      sign[0][fgs_zigzag[a*TSIZE+b]] = SIGN0(bitplane[hIndex*TSIZE+a][wIndex*TSIZE+b]);
      sign[1][fgs_zigzag[a*TSIZE+b]] = SIGN0(bitplane[hIndex*TSIZE+a][(wIndex+1)*TSIZE+b]);
      sign[2][fgs_zigzag[a*TSIZE+b]] = SIGN0(bitplane[(hIndex+1)*TSIZE+a][wIndex*TSIZE+b]);
      sign[3][fgs_zigzag[a*TSIZE+b]] = SIGN0(bitplane[(hIndex+1)*TSIZE+a][(wIndex+1)*TSIZE+b]);
    }
  }
}

/*!
 ***********************************************************************
 * \brief
 *    encodes a bitplane
 * \param pBitstream
 *    pointer to the bitstream struct
 * \param level
 *    level being decoded
 * \param level
 *    maximum level of bitplanes coded
 * \return
 *    none
 ***********************************************************************
 */

void encodeBitplane(Bitstream *pBitstream,int level,int max_level)
{
  int a,b,c,d,i,j;
  byte **yBuffer_ac,**uBuffer_ac,**vBuffer_ac;
  byte *yBuffer_dc,*uBuffer_dc,*vBuffer_dc;
  enh_block **ldata_ac = pFGSData->block_data_lum_ac;
  enh_block **udata_ac = pFGSData->block_data_u_ac;
  enh_block **vdata_ac = pFGSData->block_data_v_ac;
  enh_block **ldata_dc = pFGSData->block_data_lum_dc;
  enh_block **udata_dc = pFGSData->block_data_u_dc;
  enh_block **vdata_dc = pFGSData->block_data_v_dc;
  int mbindex,cbpindex;
  
  //allocate temporary buffers
  get_mem2D(&yBuffer_ac,16,TSIZE2);
  get_mem2D(&uBuffer_ac,4,TSIZE2);
  get_mem2D(&vBuffer_ac,4,TSIZE2);
  
  yBuffer_dc = (byte *) calloc(TSIZE2, sizeof(byte));
  uBuffer_dc = (byte *) calloc(4, sizeof(byte));
  vBuffer_dc = (byte *) calloc(4, sizeof(byte));

  mbindex = getMBIndex(pFGSData->maximum_level-level);
  //encode luminance DC bitplane
  
  i = 0;
  for(a=0; a<pFGSData->hDCTMBlocks; a++)
  {
    for(b=0; b<pFGSData->wDCTMBlocks; b++)
    {
      //scanning of the DC values
      for(c=0; c<TSIZE; c++)
      {
        for(d=0; d<TSIZE; d++)
        {
          yBuffer_dc[fgs_zigzag[c*TSIZE+d]] = pFGSData->YBitplane_dc[a*TSIZE+c][b*TSIZE+d];
        }
      }
      //encoding of the DC values
      if(pFGSData->skip_y_dc == FALSE)
      {
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"Lum DC block %d at level %d\n",i,level);
          fflush(pFGSData->pFGSTraceFile);
        }
#endif
        rleEncoding(yBuffer_dc,ldata_dc[i],0);
        encodeBlockBP_lumDC(ldata_dc[i],level);
      }
      i++;
    }
  }
  //do not encode more (there is only DC luminance values in this bitplane)
  if(mbindex == -1){
    free_mem2D(yBuffer_ac);
    free_mem2D(uBuffer_ac);
    free_mem2D(vBuffer_ac);
    free(yBuffer_dc);
    free(uBuffer_dc);
    free(vBuffer_dc);
    return;
  }

  i = 0;
  for(a=0; a<pFGSData->hDCTMBlocks; a++)
  {
    for(b=0; b<pFGSData->wDCTMBlocks; b++)
    {
      //write the CBP index for the macroblock
      //only for the MSB and MSB-1
      if(level == 0 || level == 1)
      {
        cbpindex = getCbpIndex(mbindex,i,pFGSData->maximum_level-level);
        if(level == 0) 
          writeCbpindexMSB(mbindex,i,cbpindex,pFGSData->except_code);
        else 
          writeCbpindex(mbindex,i,cbpindex,pFGSData->except_code);
      }
      //zig-zag scan of the DCT coeficients
      zigzag_scan(4*a,4*b,0,yBuffer_ac,pFGSData->YBitplane_ac);
      zigzag_scan(4*a,4*b+2,4,yBuffer_ac,pFGSData->YBitplane_ac);
      zigzag_scan(4*a+2,4*b,8,yBuffer_ac,pFGSData->YBitplane_ac);
      zigzag_scan(4*a+2,4*b+2,12,yBuffer_ac,pFGSData->YBitplane_ac);

      zigzag_scan(2*a,2*b,0,uBuffer_ac,pFGSData->UBitplane_ac);
      zigzag_scan(2*a,2*b,0,vBuffer_ac,pFGSData->VBitplane_ac);

      for(c=0; c<TSIZE; c++)
      {
        uBuffer_dc[c] = pFGSData->UBitplane_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c];
        vBuffer_dc[c] = pFGSData->VBitplane_dc[a*TSIZE*pFGSData->wDCTMBlocks+b*TSIZE+c];
      }
      //run lenght encoding
      for(j=0;j < 4; j++)
      {
        rleEncoding(yBuffer_ac[j],ldata_ac[i],j);
        rleEncoding(yBuffer_ac[4+j],ldata_ac[i+1],j);
        rleEncoding(yBuffer_ac[8+j],ldata_ac[i+2],j);
        rleEncoding(yBuffer_ac[12+j],ldata_ac[i+3],j);
      }
      rleEncoding2x2(uBuffer_dc,udata_dc[i/4],0);
      rleEncoding2x2(vBuffer_dc,vdata_dc[i/4],0);
      for(j=0;j < 4; j++)
      {
        rleEncoding(uBuffer_ac[j],udata_ac[i/4],j);
        rleEncoding(vBuffer_ac[j],vdata_ac[i/4],j);
      }
      //entropy encoding
      if(pFGSData->skip_y_ac == FALSE)
      {
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"Lum AC block %d at level %d\n",i,level);
          fflush(pFGSData->pFGSTraceFile);
        }
#endif
        encodeBlockBP(ldata_ac[i],level,max_level);
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"Lum AC block %d at level %d\n",i+1,level);
          fflush(pFGSData->pFGSTraceFile);
        }
#endif
        encodeBlockBP(ldata_ac[i+1],level,max_level);
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"Lum AC block %d at level %d\n",i+2,level);
          fflush(pFGSData->pFGSTraceFile);
        }
#endif
        encodeBlockBP(ldata_ac[i+2],level,max_level);
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"Lum AC block %d at level %d\n",i+3,level);
          fflush(pFGSData->pFGSTraceFile);
        }
#endif
        encodeBlockBP(ldata_ac[i+3],level,max_level);
      }
      if(pFGSData->skip_chr_dc == FALSE || pFGSData->skip_chr_ac == FALSE)
      {
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"Chr U DC block %d at level %d\n",i/4,level);
          fflush(pFGSData->pFGSTraceFile);    
        }
#endif
        encodeBlockBP_chrDC(udata_dc[i/4],level,max_level);
#ifdef FGS_TRACE        
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"Chr V DC block %d at level %d\n",i/4,level);
          fflush(pFGSData->pFGSTraceFile);
        }
#endif        
        encodeBlockBP_chrDC(vdata_dc[i/4],level,max_level);
      }
      if(pFGSData->skip_chr_ac == FALSE)
      {
#ifdef FGS_TRACE        
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"Chr AC block %d at level %d\n",i/4,level);
          fflush(pFGSData->pFGSTraceFile);
        }
#endif
        encodeBlockBP(udata_ac[i/4],level,max_level);
#ifdef FGS_TRACE       
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"Chr AC block %d at level %d\n",i/4,level);
          fflush(pFGSData->pFGSTraceFile);
        }
#endif
        encodeBlockBP(vdata_ac[i/4],level,max_level);
      }
      i+=4;
    }
  }
  free_mem2D(yBuffer_ac);
  free_mem2D(uBuffer_ac);
  free_mem2D(vBuffer_ac);
  free(yBuffer_dc);
  free(uBuffer_dc);
  free(vBuffer_dc);
  return;
}

void zigzag_scan(int hIndex,int wIndex,int i,byte **buffer,byte **bitplane)
{
  int a,b;

  //scanning in zigzag order the bitplane to store in a lum & chrominance buffer
  for(a=0; a<TSIZE; a++)
  {
    for(b=0; b<TSIZE; b++)
    {
      buffer[i][fgs_zigzag[a*TSIZE+b]] = bitplane[hIndex*TSIZE+a][wIndex*TSIZE+b];
      buffer[i+1][fgs_zigzag[a*TSIZE+b]] = bitplane[hIndex*TSIZE+a][(wIndex+1)*TSIZE+b];
      buffer[i+2][fgs_zigzag[a*TSIZE+b]] = bitplane[(hIndex+1)*TSIZE+a][wIndex*TSIZE+b];
      buffer[i+3][fgs_zigzag[a*TSIZE+b]] = bitplane[(hIndex+1)*TSIZE+a][(wIndex+1)*TSIZE+b];
    }
  }
}

/*!
 ***********************************************************************
 * \brief
 *    encodes a block BP
 * \param block_data
 *    pointer to enhancement block data
 * \param level
 *    level being decoded
 * \param level
 *    maximum level of bitplanes coded
 * \return
 *    none
 ***********************************************************************
 */

void encodeBlockBP_lumDC(enh_block* block_data, int level)
{
  int i,n,os,ind_table,ind_level,data,len;

  n = 0; os=0;

  if(level > 2)
  {
    ind_level = 2;
  }
  else
  {
    ind_level = level;
  }

  if(block_data->run[0][0]==15 && block_data->end[0][0]==0) //all-zero case
  {
    data = table_dc_lum_uvlc[2*ind_level][block_data->run[0][0]+block_data->end[0][0]*16];
    len = table_dc_lum_uvlc[2*ind_level+1][block_data->run[0][0]+block_data->end[0][0]*16];
    writeIntFGS(data,len,pFGSBitstream);
    
#ifdef FGS_TRACE
    if(gtr == 146)
    {
      fprintf(pFGSData->pFGSTraceFile,"RLE - Run:%d Lenght:%d\n",block_data->run[0][0],block_data->end[0][0]);
      fflush(pFGSData->pFGSTraceFile); 
    }
#endif
  }
  else 
  {
    while (block_data->end[0][n] == 0) n += 1;
    n += 1;
  }
  for(i = 0;i < n;i++)
  {
    ind_table = block_data->run[0][i];
    data = table_dc_lum_uvlc[2*ind_level][block_data->run[0][i]+block_data->end[0][i]*16];
    len = table_dc_lum_uvlc[2*ind_level+1][block_data->run[0][i]+block_data->end[0][i]*16];
    writeIntFGS(data,len,pFGSBitstream);

#ifdef FGS_TRACE
    if(gtr == 146)
    {
      fprintf(pFGSData->pFGSTraceFile,"RLE - Run:%d Lenght:%d\n",block_data->run[0][i],block_data->end[0][i]);
      fflush(pFGSData->pFGSTraceFile); 
    }
#endif
    if (block_data->sign[0][os+ind_table]!=2 && block_data->sign[0][os+ind_table]!=0)
    {
#ifdef FGS_TRACE
      if(gtr == 146)
      {
        fprintf(pFGSData->pFGSTraceFile,"Signal:%d\n",block_data->sign[0][os+ind_table]==-1?1:0);
        fflush(pFGSData->pFGSTraceFile); 
      }
#endif
      writeIntFGS((block_data->sign[0][os+ind_table]==-1?1:0),1,pFGSBitstream);
      block_data->sign[0][os+ind_table] = 2;
    }
    os += block_data->run[0][i]+1;
  }
}

/*!
 ***********************************************************************
 * \brief
 *    encodes a block BP
 * \param block_data
 *    pointer to enhancement block data
 * \param level
 *    level being decoded
 * \param level
 *    maximum level of bitplanes coded
 * \return
 *    none
 ***********************************************************************
 */

void encodeBlockBP_chrDC(enh_block* block_data, int level, int max_level)
{
  int i,n,ind_table,os,temp,ind_level,data,len;

  if(block_data->cbp != 1 && level > 1)
  {
    //if the MSB bitplane of the block is reached in this level or in one below
    if(block_data->data_level >= (max_level-level))
    {
      //write '0'
      writeByteFGS(0x00,1,pFGSBitstream);
      block_data->cbp = 1;
    }
    else
    {
      //otherwise write '1'
      writeByteFGS(0x01,1,pFGSBitstream);
      block_data->cbp = 0;
    }
  }
  n = 0;
  if(block_data->cbp == 1)
  {
    temp = block_data->data_level - (max_level-level);
    
    if(temp > 2)
    {
      ind_level = 2;
    }
    else
    {
      ind_level = temp;
    }

    if(block_data->run[0][0]==3 && block_data->end[0][0]==0) //all-zero case
    {
      data = table_dc_chr_uvlc[2*ind_level][block_data->run[0][0]+block_data->end[0][0]*4];
      len = table_dc_chr_uvlc[2*ind_level+1][block_data->run[0][0]+block_data->end[0][0]*4];
      writeIntFGS(data,len,pFGSBitstream);

#ifdef FGS_TRACE
      if(gtr == 146)
      {
        fprintf(pFGSData->pFGSTraceFile,"RLE - Run:%d Lenght:%d\n",block_data->run[0][0],block_data->end[0][0]);
        fflush(pFGSData->pFGSTraceFile); 
      }
#endif
    }
    else 
    {
      while (block_data->end[0][n] == 0) n += 1;
      n += 1;
    }
    os = 0;
    for(i = 0;i < n;i++)
    {
      ind_table = block_data->run[0][i];
      
      data = table_dc_chr_uvlc[2*ind_level][block_data->run[0][i]+block_data->end[0][i]*4];
      len = table_dc_chr_uvlc[2*ind_level+1][block_data->run[0][i]+block_data->end[0][i]*4];
      writeIntFGS(data,len,pFGSBitstream);
      
#ifdef FGS_TRACE
      if(gtr == 146)
      {
        fprintf(pFGSData->pFGSTraceFile,"RLE - Run:%d Lenght:%d\n",block_data->run[0][i],block_data->end[0][i]);
        fflush(pFGSData->pFGSTraceFile); 
      }
#endif
      if (block_data->sign[0][os+ind_table]!=2 && block_data->sign[0][os+ind_table]!=0)
      {
#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"Signal:%d\n",block_data->sign[0][os+ind_table]==-1?1:0);
          fflush(pFGSData->pFGSTraceFile); 
        }
#endif
        writeIntFGS((block_data->sign[0][os+ind_table]==-1?1:0),1,pFGSBitstream);
        block_data->sign[0][os+ind_table] = 2;
      }
      os += block_data->run[0][i]+1;
    }
  }
}
/*!
 ***********************************************************************
 * \brief
 *    encodes a block BP
 * \param block_data
 *    pointer to enhancement block data
 * \param level
 *    level being decoded
 * \param level
 *    maximum level of bitplanes coded
 * \return
 *    none
 ***********************************************************************
 */

void encodeBlockBP(enh_block* block_data, int level, int max_level)
{
  int i,index,n,os,ind_table,temp,ind_level,data,len;
  // if fgs_vop_bp_id>=2 && previous_fgs_msb_not_reached == 1
  if(block_data->cbp != 1 && level > 1)
  {
    //if the MSB bitplane of the block is reached in this level or in one below
    if(block_data->data_level >= (max_level-level))    
    {
      //write '0'
      writeByteFGS(0x00,1,pFGSBitstream);
      block_data->cbp = 1;
    }
    else
    {
      //otherwise write '1'
      writeByteFGS(0x01,1,pFGSBitstream);
      block_data->cbp = 0;
    }
  }
  if(block_data->cbp == 1)
  {
    temp = block_data->data_level - (max_level-level);
    
    if(temp > 2)
    {
      ind_level = 2;
    }
    else
    {
      ind_level = temp;
    }
    for(index = 0;index < 4;index++)
    {
      n = 0;
      if(block_data->run[index][0]==15 && block_data->end[index][0]==0) //all-zero case
      {
        data = table_ac_uvlc[2*ind_level][block_data->run[index][0]+block_data->end[index][0]*16];
        len = table_ac_uvlc[2*ind_level+1][block_data->run[index][0]+block_data->end[index][0]*16];
        writeIntFGS(data,len,pFGSBitstream);

#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"RLE - Run:%d Lenght:%d\n",block_data->run[index][0],block_data->end[index][0]);
          fflush(pFGSData->pFGSTraceFile); 
        }
#endif
      }
      else 
      {
        while (block_data->end[index][n] == 0) n += 1;
        n += 1;
      }
      os = 0;
      for(i = 0;i < n;i++)
      {
        ind_table = block_data->run[index][i];
        data = table_ac_uvlc[2*ind_level][block_data->run[index][i]+block_data->end[index][i]*16];
        len = table_ac_uvlc[2*ind_level+1][block_data->run[index][i]+block_data->end[index][i]*16];
        writeIntFGS(data,len,pFGSBitstream);

#ifdef FGS_TRACE
        if(gtr == 146)
        {
          fprintf(pFGSData->pFGSTraceFile,"RLE - Run:%d Lenght:%d\n",block_data->run[index][i],block_data->end[index][i]);
          fflush(pFGSData->pFGSTraceFile); 
        }
#endif
        if (block_data->sign[index][os+ind_table]!=2 && block_data->sign[index][os+ind_table]!=0)
        {

#ifdef FGS_TRACE
          if(gtr == 146)
          {
            fprintf(pFGSData->pFGSTraceFile,"Signal:%d\n",block_data->sign[index][os+ind_table]==-1?1:0);
            fflush(pFGSData->pFGSTraceFile); 
          }
#endif
          writeIntFGS((block_data->sign[index][os+ind_table]==-1?1:0),1,pFGSBitstream);
          block_data->sign[index][os+ind_table] = 2;
        }
        os += block_data->run[index][i]+1;
      }
    }
  }
}

/*!
 ***********************************************************************
 * \brief
 *    gets the MB index that indicates the components coded
 * \param level
 *    level being decoded
 * \return
 *    MB index
 ***********************************************************************
 */

int getMBIndex(int level)
{
  int index,yy_ac,yy_dc,dc_chroma,ac_chroma;
  yy_ac = pFGSData->maximum_level_y_ac;
  yy_dc = pFGSData->maximum_level_y_dc;
  dc_chroma = pFGSData->maximum_level_chr_dc;
  ac_chroma = pFGSData->maximum_level_chr_ac;
  
  if (yy_dc >= level && yy_ac < level && dc_chroma < level && ac_chroma < level)  //only dc coeficients are coded in this vop-bp
  {
    index = -1;
  }
  else if (yy_ac >= level && dc_chroma < level && ac_chroma < level)  //nc = 0
  {
    index = 0;
  }
  else if (yy_ac >= level && dc_chroma >= level && ac_chroma < level)  //nc = 1
  {
    index = 1;
  }
  else if (yy_ac >= level && ac_chroma >= level) //nc = 2
  {
    index = 2;
  }
  else if (yy_ac < level && dc_chroma >= level && ac_chroma < level) //nc = 1
  {
    index = 3;
  }
  else if (yy_ac < level && ac_chroma >= level) //nc = 2
  {
    index = 4;
  }
  else
  {
    printf("Error in the getMBIndex\n");
    exit(1);
  }
  return(index);
}

/*!
 ***********************************************************************
 * \brief
 *    gets the cbp_pattern from the data_level values (calculated for each block)
 * \param comb_index
 *    index to indicate which components are coded
 * \param block_index
 *    index to indicate the number of the block being processed
 * \param level
 *    level being decoded
 * \return
 *    cbp_pattern
 ***********************************************************************
 */

int getCbpIndex(int mbIndex,int blockIndex,int level)
{
  int cbp_pattern, i, temp[8], temp1[6], nc, max_level,max_level1;

  for (i=0; i < 4; i++)
  {
    temp[i] = pFGSData->block_data_lum_ac[blockIndex+i]->data_level - level;
  }
  temp[4] = pFGSData->block_data_u_dc[blockIndex/4]->data_level - level;
  temp[5] = pFGSData->block_data_v_dc[blockIndex/4]->data_level - level;
  temp[6] = pFGSData->block_data_u_ac[blockIndex/4]->data_level - level;
  temp[7] = pFGSData->block_data_v_ac[blockIndex/4]->data_level - level;

  for (i=0; i < 4; i++)
  {
    if(temp[i] >= 0) temp1[i] = 1;  //encode
    else temp1[i] = 0;              //do not encode
  }
  if(temp[4] < 0 && temp[5] < 0 && temp[6] < 0 && temp[7] < 0) nc = 0;
  else if(temp[6] < 0 && temp[7] < 0) nc = 1;
  else nc = 2;

  cbp_pattern = 0;

  switch(mbIndex) 
  {
    case 0: 
      cbp_pattern = (temp1[0] << 3) + (temp1[1] << 2) + (temp1[2] << 1) + temp1[3];
      break;
    case 1: 
    case 2:
      cbp_pattern = (nc << 4) + (temp1[0] << 3) + (temp1[1] << 2) + (temp1[2] << 1) + temp1[3];
      break;
    case 3:
    case 4: 
      cbp_pattern = nc;
      break;
    default:
      printf("Error in the getCbpIndex!\n");
      exit(1);
  }
  //data_levels actualization due to the way of cbp is coded 
  if(nc == 1)
  {
    max_level = max(pFGSData->block_data_u_dc[blockIndex/4]->data_level,
      pFGSData->block_data_v_dc[blockIndex/4]->data_level);
    pFGSData->block_data_u_dc[blockIndex/4]->data_level = max_level;
    pFGSData->block_data_v_dc[blockIndex/4]->data_level = max_level;
  }
  else if(nc == 2)
  {
    max_level = max(pFGSData->block_data_u_dc[blockIndex/4]->data_level,
      pFGSData->block_data_v_dc[blockIndex/4]->data_level);
    max_level1 = max(pFGSData->block_data_u_ac[blockIndex/4]->data_level,
      pFGSData->block_data_v_ac[blockIndex/4]->data_level);
    max_level = max(max_level,max_level1);

    pFGSData->block_data_u_dc[blockIndex/4]->data_level = max_level;
    pFGSData->block_data_v_dc[blockIndex/4]->data_level = max_level;
    pFGSData->block_data_u_ac[blockIndex/4]->data_level = max_level1;
    pFGSData->block_data_v_ac[blockIndex/4]->data_level = max_level1;
  }
  return(cbp_pattern);
}

/*!
 ***********************************************************************
 * \brief
 *    writes the cbp (code bit pattern) of fgs_msb_not_reached for the MSB vop-bp
 * \param comb_index
 *    index to indicate which components are coded
 * \param block_index
 *    index to indicate the number of the block being processed
 * \param cbp_index
 *    index to indicate the pattern to write
 * \notes 
 *    except_code must be zeroed...
 * \return
 *    none
 ***********************************************************************
 */

void writeCbpindexMSB(int comb_index,int block_index,int cbp_index, byte *except_code)
{
  int data,len;
  enh_block **ldata_ac = pFGSData->block_data_lum_ac;
  enh_block **udata_ac = pFGSData->block_data_u_ac;
  enh_block **vdata_ac = pFGSData->block_data_v_ac;
  enh_block **udata_dc = pFGSData->block_data_u_dc;
  enh_block **vdata_dc = pFGSData->block_data_v_dc;

#ifdef FGS_TRACE
  if(gtr == 146)
  {
    fprintf(pFGSData->pFGSTraceFile,"CombIndex: %d  CbpIndex: %d\n",comb_index,cbp_index);
    fflush(pFGSData->pFGSTraceFile); 
  }
#endif
 
  if(cbp_index == 0) except_code[block_index/4] = 1;
  switch(comb_index) 
  {
  case 0:
    data = table_cbp_yyyy[0][cbp_index];
    len = table_cbp_yyyy[1][cbp_index];
    writeIntFGS(data,len,pFGSBitstream);
    ldata_ac[block_index]->cbp = 0x1 & (cbp_index >> 3);
    ldata_ac[block_index+1]->cbp = 0x1 & (cbp_index >> 2);
    ldata_ac[block_index+2]->cbp = 0x1 & (cbp_index >> 1);
    ldata_ac[block_index+3]->cbp = 0x1 & cbp_index;
  break;
  case 1:
  case 2:
    data = table_cbp_yyyyuv[0][cbp_index];
    len = table_cbp_yyyyuv[1][cbp_index];
    writeIntFGS(data,len,pFGSBitstream);
    ldata_ac[block_index]->cbp = 0x1 & (cbp_index >> 3);
    ldata_ac[block_index+1]->cbp = 0x1 & (cbp_index >> 2);
    ldata_ac[block_index+2]->cbp = 0x1 & (cbp_index >> 1);
    ldata_ac[block_index+3]->cbp = 0x1 & cbp_index;
    if((cbp_index>>4) == 0)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x00;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x00;
    }
    if((cbp_index>>4) == 1)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    if((cbp_index>>4) == 2)
    {
      udata_ac[block_index/4]->cbp = 0x01;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x01;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
  break;
  case 3:
  case 4:
    if(cbp_index == 0)
    {
      writeIntFGS(1,1,pFGSBitstream);
    }
    else
    {
      writeIntFGS(0,1,pFGSBitstream);
      writeIntFGS(cbp_index,2,pFGSBitstream);
    }
    if(cbp_index == 0)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x00;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x00;
    }
    if(cbp_index == 1)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    if(cbp_index == 2)
    {
      udata_ac[block_index/4]->cbp = 0x01;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x01;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
  break;
  default:
    printf("Error in the writeCbpIndexMSB!\n");
    exit(1);
  }
}

/*!
 ***********************************************************************
 * \brief
 *    writes the cbp (code bit pattern) of fgs_msb_not_reached for the second vop-bp
 * \param comb_index
 *    index to indicate which components are coded
 * \param block_index
 *    index to indicate the number of the block being processed
 * \param cbp_index
 *    index to indicate the pattern to write
 * \return
 *    none
 ***********************************************************************
 */

void writeCbpindex(int comb_index,int block_index,int cbp_index,byte *except_code)
{
  int data,len;
  enh_block **ldata_ac = pFGSData->block_data_lum_ac;
  enh_block **udata_ac = pFGSData->block_data_u_ac;
  enh_block **vdata_ac = pFGSData->block_data_v_ac;
  enh_block **udata_dc = pFGSData->block_data_u_dc;
  enh_block **vdata_dc = pFGSData->block_data_v_dc;

#ifdef FGS_TRACE
  if(gtr == 146)
  {
    fprintf(pFGSData->pFGSTraceFile,"CombIndex: %d  CbpIndex: %d  except_code: %d\n",comb_index,cbp_index,except_code[block_index/4]);
    fflush(pFGSData->pFGSTraceFile); 
  }
#endif

  if(except_code[block_index/4] == 1)
  {    
    switch(comb_index)
    {
    case 0:
      data = table_cbp_yyyy[2][cbp_index];
      len = table_cbp_yyyy[3][cbp_index];
      writeIntFGS(data,len,pFGSBitstream);
      ldata_ac[block_index]->cbp = 0x1 & (cbp_index >> 3);
      ldata_ac[block_index+1]->cbp = 0x1 & (cbp_index >> 2);
      ldata_ac[block_index+2]->cbp = 0x1 & (cbp_index >> 1);
      ldata_ac[block_index+3]->cbp = 0x1 & cbp_index;
      break;
    case 1:
    case 2:
      data = table_cbp_yyyyuv[2][cbp_index];
      len = table_cbp_yyyyuv[3][cbp_index];
      writeIntFGS(data,len,pFGSBitstream);
      ldata_ac[block_index]->cbp = 0x1 & (cbp_index >> 3);
      ldata_ac[block_index+1]->cbp = 0x1 & (cbp_index >> 2);
      ldata_ac[block_index+2]->cbp = 0x1 & (cbp_index >> 1);
      ldata_ac[block_index+3]->cbp = 0x1 & cbp_index;
      if((cbp_index>>4) == 0)
      {
        udata_ac[block_index/4]->cbp = 0x00;
        udata_dc[block_index/4]->cbp = 0x00;
        vdata_ac[block_index/4]->cbp = 0x00;
        vdata_dc[block_index/4]->cbp = 0x00;
      }
      if((cbp_index>>4) == 1)
      {
        udata_ac[block_index/4]->cbp = 0x00;
        udata_dc[block_index/4]->cbp = 0x01;
        vdata_ac[block_index/4]->cbp = 0x00;
        vdata_dc[block_index/4]->cbp = 0x01;
      }
      if((cbp_index>>4) == 2)
      {
        udata_ac[block_index/4]->cbp = 0x01;
        udata_dc[block_index/4]->cbp = 0x01;
        vdata_ac[block_index/4]->cbp = 0x01;
        vdata_dc[block_index/4]->cbp = 0x01;
      }
      break;
    case 3:
    case 4:
      writeIntFGS(cbp_index,2,pFGSBitstream);
      if(cbp_index == 0)
      {
        udata_ac[block_index/4]->cbp = 0x00;
        udata_dc[block_index/4]->cbp = 0x00;
        vdata_ac[block_index/4]->cbp = 0x00;
        vdata_dc[block_index/4]->cbp = 0x00;
      }
      if(cbp_index == 1)
      {
        udata_ac[block_index/4]->cbp = 0x00;
        udata_dc[block_index/4]->cbp = 0x01;
        vdata_ac[block_index/4]->cbp = 0x00;
        vdata_dc[block_index/4]->cbp = 0x01;
      }
      if(cbp_index == 2)
      {
        udata_ac[block_index/4]->cbp = 0x01;
        udata_dc[block_index/4]->cbp = 0x01;
        vdata_ac[block_index/4]->cbp = 0x01;
        vdata_dc[block_index/4]->cbp = 0x01;
      }
      break;
    default:
      printf("Error in the writeCbpIndexMSB!\n");
      exit(1);
    }
    return;
  }
  switch(comb_index)
  {
  case 0:
    data = table_cbp_yyyy[4][cbp_index];
    len = table_cbp_yyyy[5][cbp_index];
    writeIntFGS(data,len,pFGSBitstream);
    ldata_ac[block_index]->cbp = 0x1 & (cbp_index >> 3);
    ldata_ac[block_index+1]->cbp = 0x1 & (cbp_index >> 2);
    ldata_ac[block_index+2]->cbp = 0x1 & (cbp_index >> 1);
    ldata_ac[block_index+3]->cbp = 0x1 & cbp_index;
    break;
  case 1:
  case 2:
    data = table_cbp_yyyyuv[4][cbp_index];
    len = table_cbp_yyyyuv[5][cbp_index];
    writeIntFGS(data,len,pFGSBitstream);
    ldata_ac[block_index]->cbp = 0x1 & (cbp_index >> 3);
    ldata_ac[block_index+1]->cbp = 0x1 & (cbp_index >> 2);
    ldata_ac[block_index+2]->cbp = 0x1 & (cbp_index >> 1);
    ldata_ac[block_index+3]->cbp = 0x1 & cbp_index;
    if((cbp_index>>4) == 0)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x00;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x00;
    }
    if((cbp_index>>4) == 1)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    if((cbp_index>>4) == 2)
    {
      udata_ac[block_index/4]->cbp = 0x01;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x01;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    break;
  case 3:
  case 4:
    writeIntFGS(cbp_index,2,pFGSBitstream);
    if(cbp_index == 0)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x00;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x00;
    }
    if(cbp_index == 1)
    {
      udata_ac[block_index/4]->cbp = 0x00;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x00;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    if(cbp_index == 2)
    {
      udata_ac[block_index/4]->cbp = 0x01;
      udata_dc[block_index/4]->cbp = 0x01;
      vdata_ac[block_index/4]->cbp = 0x01;
      vdata_dc[block_index/4]->cbp = 0x01;
    }
    break;
  default:
    printf("Error in the writeCbpIndexMSB!\n");
    exit(1);
  }  
}

/*!
 ***********************************************************************
 * \brief
 *    returns the number of ones in a integer
 * \param number
 *    input number
 * \return
 *    number of ones
 ***********************************************************************
 */

int nr_ones(int number)
{
  int i, ones;
  ones = 0;
  for (i=0; i < 31; i++) 
  {
    ones += 0x1 & (number>>i);
  }
  return(ones);
}

/*!
 ***********************************************************************
 * \brief
 *    performs the run lenght encoding in a block
 * \param in
 *    pointer to input buffer
 * \param block_data
 *    pointer to enh_block structure
 * \param level
 *    bitplane level
 * \return
 *    none
 ***********************************************************************
 */

void rleEncoding(byte *in, enh_block* block_data, int t_block)
{
  int i;
  int n;
  int curr_run;
  int runs[TSIZE2];

  curr_run = 0;
  n = 0;
  for (i=0;i < TSIZE2;i++) 
  {
    if (in[i] == 0) 
    {
      curr_run += 1;
    } 
    else 
    {
      runs[n] = curr_run;
      curr_run = 0;
      n++;
    }
  }
  if(n == 0) 
  {
    block_data->run[t_block][0] = 15;
    block_data->end[t_block][0] = 0;
  }
  for(i=0;i<n;i++)
  {
    block_data->run[t_block][i] = runs[i];

    if(i != n-1) 
    {
      block_data->end[t_block][i] = 0;
    } 
    else 
    {
      block_data->end[t_block][i] = 1;
    }
  }
}

void rleEncoding2x2(byte *in, enh_block* block_data, int t_block)
{
  int i;
  int n;
  int curr_run;
  int runs[4];

  curr_run = 0;
  n = 0;
  for (i=0;i < 4;i++) 
  {
    if (in[i] == 0) 
    {
      curr_run += 1;
    } 
    else 
    {
      runs[n] = curr_run;
      curr_run = 0;
      n++;
    }
  }
  if(n == 0) 
  {
    block_data->run[t_block][0] = 3;
    block_data->end[t_block][0] = 0;
  }
  for(i=0;i<n;i++)
  {
    block_data->run[t_block][i] = runs[i];

    if(i != n-1) 
    {
      block_data->end[t_block][i] = 0;
    } 
    else 
    {
      block_data->end[t_block][i] = 1;
    }
  }
}

/*!
 ***********************************************************************
 * \brief
 *    free memory used for FGS encoding
 * \para Notes
 *    beware with the structure
 * \return
 *    none
 ***********************************************************************
 */

void cleanFGS()
{
  int a,nblocks;
  //writes the code to signal end of sequence
  a = VISUAL_OBJECT_SEQUENCE_END_CODE;
  fwrite(&a,sizeof(byte),VISUAL_OBJECT_SEQUENCE_END_CODE_LENGHT,pFGSData->pFGSFile);
  //closes the bitstream file
  fclose(pFGSData->pFGSFile);
  //closes the stat trace file
  fclose(pFGSData->pFGSTraceFile);
  //free residual images
  free_mem2Dint(pFGSData->Ydiff_ac);
  free_mem2Dint(pFGSData->Udiff_ac);
  free_mem2Dint(pFGSData->Vdiff_ac);
  free_mem2Dint(pFGSData->Ydiff_dc);
  free(pFGSData->Udiff_dc);
  free(pFGSData->Vdiff_dc);
  //free bitplanes
  free_mem2D(pFGSData->YBitplane_ac);
  free_mem2D(pFGSData->UBitplane_ac);
  free_mem2D(pFGSData->VBitplane_ac);
  free_mem2D(pFGSData->YBitplane_dc);
  free(pFGSData->UBitplane_dc);
  free(pFGSData->VBitplane_dc);

  //free structure with encoding data for luminance
  nblocks = pFGSData->hDCTMBlocks*pFGSData->wDCTMBlocks;
  for(a=0;a < (4*nblocks);a++)
  {
    free(pFGSData->block_data_lum_ac[a]);
  }
  free(pFGSData->block_data_lum_ac);

  for(a=0;a < nblocks;a++)
  {
    free(pFGSData->block_data_lum_dc[a]);
  }
  free(pFGSData->block_data_lum_dc);

  //free structure with encoding data for chrominance
  nblocks = pFGSData->hDCTBlocks_cr*pFGSData->wDCTBlocks_cr;
  for(a=0;a < nblocks;a++)
  {
    free(pFGSData->block_data_u_ac[a]);
    free(pFGSData->block_data_v_ac[a]);
    free(pFGSData->block_data_u_dc[a]);
    free(pFGSData->block_data_v_dc[a]);
  }
  free(pFGSData->block_data_u_ac);
  free(pFGSData->block_data_v_ac);
  free(pFGSData->block_data_u_dc);
  free(pFGSData->block_data_v_dc);

  //free space for exception code
  free(pFGSData->except_code);

  //free space for fgs encoding data
  free(pFGSData);
  //free bitstream buffer
  free(pFGSBitstream->streamBuffer);
  free(pFGSBitstream);
}

/*!
 ***********************************************************************
 * \brief
 *    performs the H.26L integer DCT in a 16x16 Macroblock for the luma component
 * \param coefAC
 *    pointer to matrix (macroblock) with 16x16 values
 * \param coefDC
 *    pointer to empty matrix
 * \return
 *    coefAC -> DCT AC coeficients
 *    coefDC -> DCT DC coeficients
 ***********************************************************************
 */

void H26L_transLum(int **coefAC, int **coefDC)
{
  int i,j,ii,jj,i1,j1,level;
  int M5[4];
  int M0[4][4][4][4];

  //put in index,block_num,index,block_num 4 dimensional array

  for (j=0;j<16;j++)
  {
    for (i=0;i<16;i++)
    {
      M0[i%4][i/4][j%4][j/4]=coefAC[i][j];
    }
  }

  for (jj=0;jj<4;jj++)
  {
    for (ii=0;ii<4;ii++)
    {
      for (j=0;j<4;j++)
      {
        for (i=0;i<2;i++)
        {
          i1=3-i;
          M5[i]=  M0[i][ii][j][jj]+M0[i1][ii][j][jj];
          M5[i1]= M0[i][ii][j][jj]-M0[i1][ii][j][jj];
        }
        M0[0][ii][j][jj]=M5[0]+M5[1];
        M0[2][ii][j][jj]=M5[0]-M5[1];
        M0[1][ii][j][jj]=M5[3]*2+M5[2];
        M0[3][ii][j][jj]=M5[3]-M5[2]*2;
      }
      // vertical
      for (i=0;i<4;i++)
      {
        for (j=0;j<2;j++)
        {
          j1=3-j;
          M5[j] = M0[i][ii][j][jj]+M0[i][ii][j1][jj];
          M5[j1]= M0[i][ii][j][jj]-M0[i][ii][j1][jj];
        }
        M0[i][ii][0][jj]=M5[0]+M5[1];
        M0[i][ii][2][jj]=M5[0]-M5[1];
        M0[i][ii][1][jj]=M5[3]*2+M5[2];
        M0[i][ii][3][jj]=M5[3] -M5[2]*2;
      }
    }
  }

  // pick out DC coeff

  for (j=0;j<4;j++)
  {
    for (i=0;i<4;i++)
    {
       coefDC[i][j]= M0[0][i][0][j];
       M0[0][i][0][j] = 0;
    }
  }

  //calulates the hadamard symetric transform

  for (j=0;j<4;j++)
  {
    for (i=0;i<2;i++)
    {
      i1=3-i;
      M5[i]= coefDC[i][j]+coefDC[i1][j];
      M5[i1]=coefDC[i][j]-coefDC[i1][j];
    }
    coefDC[0][j]=M5[0]+M5[1];
    coefDC[2][j]=M5[0]-M5[1];
    coefDC[1][j]=M5[3]+M5[2];
    coefDC[3][j]=M5[3]-M5[2];
  }

  // vertical

  for (i=0;i<4;i++)
  {
    for (j=0;j<2;j++)
    {
      j1=3-j;
      M5[j]= coefDC[i][j]+coefDC[i][j1];
      M5[j1]=coefDC[i][j]-coefDC[i][j1];
    }
    coefDC[i][0]=(M5[0]+M5[1])/2;
    coefDC[i][2]=(M5[0]-M5[1])/2;
    coefDC[i][1]=(M5[3]+M5[2])/2;
    coefDC[i][3]=(M5[3]-M5[2])/2;
  }

  //scaling of the AC coeficients (Q=0) and put back the AC transformed coeficients

  for (j=0;j<16;j++)
  {
    for (i=0;i<16;i++)
    {
       level = (abs(M0[i%4][i/4][j%4][j/4])*fgs_quant_coef[i%4][j%4]+(1<<16)/3) >> 16;
       coefAC[i][j] = sign(level,M0[i%4][i/4][j%4][j/4]);
    }
  }

  //scaling of the DC coeficients (Q=0)
  for (j=0;j<4;j++)
  {
    for (i=0;i<4;i++)
    {
      level = (abs(coefDC[i][j])*fgs_quant_coef[0][0]+2*(1<<16)/3) >> 17;
      coefDC[i][j] = sign(level,coefDC[i][j]);
    }
  }
}

/*!
 ***********************************************************************
 * \brief
 *    performs the H.26L integer DCT in a 8x8 block for the chroma components
 * \param data
 *    pointer to matrix with 8x8 values
 * \return
 *    none
 ***********************************************************************
 */
void H26L_transChr(int ** coefAC,int *coefDC)
{
  int i,j,ii,jj,i1,j1,level;
  int M5[4];
  int M0[4][2][4][2];
  int M1[4][4];

  //put in block_numX,indexX,block_numY,indexY 4th array

  for (j=0;j<8;j++)
  {
    for (i=0;i<8;i++)
    {
      M0[i%4][i/4][j%4][j/4]=coefAC[i][j];
    }
  }

  for (jj=0;jj<2;jj++)
  {
    for (ii=0;ii<2;ii++)
    {
      for (j=0;j<4;j++)
      {
        for (i=0;i<2;i++)
        {
          i1=3-i;
          M5[i]=  M0[i][ii][j][jj]+M0[i1][ii][j][jj];
          M5[i1]= M0[i][ii][j][jj]-M0[i1][ii][j][jj];
        }
        M0[0][ii][j][jj]=M5[0]+M5[1];
        M0[2][ii][j][jj]=M5[0]-M5[1];
        M0[1][ii][j][jj]=M5[3]*2+M5[2];
        M0[3][ii][j][jj]=M5[3]-M5[2]*2;
      }
      // vertical
      for (i=0;i<4;i++)
      {
        for (j=0;j<2;j++)
        {
          j1=3-j;
          M5[j] = M0[i][ii][j][jj]+M0[i][ii][j1][jj];
          M5[j1]= M0[i][ii][j][jj]-M0[i][ii][j1][jj];
        }
        M0[i][ii][0][jj]=M5[0]+M5[1];
        M0[i][ii][2][jj]=M5[0]-M5[1];
        M0[i][ii][1][jj]=M5[3]*2+M5[2];
        M0[i][ii][3][jj]=M5[3]-M5[2]*2;
      }
    }
  }

  // pick out DC coeff
  for (j=0;j<2;j++)
  {
    for (i=0;i<2;i++)
    {
      M1[i][j]= M0[0][i][0][j];
      M0[0][i][0][j] = 0;
    }
  }

  //hadamard transform of the DC (2x2) coeficients
  coefDC[0]=(M1[0][0]+M1[1][0]+M1[0][1]+M1[1][1]);
  coefDC[1]=(M1[0][0]-M1[1][0]+M1[0][1]-M1[1][1]);
  coefDC[2]=(M1[0][0]+M1[1][0]-M1[0][1]-M1[1][1]);
  coefDC[3]=(M1[0][0]-M1[1][0]-M1[0][1]+M1[1][1]);


  // Scaling of the DC coeficients
  
  for (i=0;i<4;i++)
  {
    level = (abs(coefDC[i])*fgs_quant_coef[0][0]+2*(1<<16)/3) >> 17;
    coefDC[i] = sign(level,coefDC[i]);
  }

  // Scaling of the AC coeficients and put back the AC transformed coeficients

  for (j=0;j<8;j++)
  {
    for (i=0;i<8;i++)
    {
      level = (abs(M0[i%4][i/4][j%4][j/4])*fgs_quant_coef[i%4][j%4]+(1<<16)/3) >> 16;
      coefAC[i][j] = sign(level,M0[i%4][i/4][j%4][j/4]);
    }
  }
}

/*!
 ***********************************************************************
 * \brief
 *    performs the forward discrete cosine transform in an image
 * \param data
 *    pointer to matrix image data
 * \param hDCTBlocks
 *    number of blocks relative to the height of the image (vertical blocks)
 * \param wDCTBlocks
 *    number of blocks relative to the width of the image (horizontal blocks)
 * \para Notes
 *    the first cycle must be speed up with memcpy
 * \return
 *    none
 ***********************************************************************
 */

void calcftrans_lum(int ** dataAC, int ** dataDC,int hDCTBlocks,int wDCTBlocks)
{
  int a,b,c,d;
  int **ACBuffer,**DCBuffer;
   //temporary buffer for MB store
  get_mem2Dint(&ACBuffer, MBSIZE, MBSIZE);
  get_mem2Dint(&DCBuffer, TSIZE, TSIZE);

  for(a=0; a<hDCTBlocks; a++)
  {
    for(b=0; b<wDCTBlocks; b++)
    {
      for(c=0; c<MBSIZE; c++)
      {
        for(d=0; d<MBSIZE; d++)
        {
          ACBuffer[c][d] = dataAC[a*MBSIZE+c][b*MBSIZE+d];
        }
      }
      H26L_transLum(ACBuffer,DCBuffer);
      for(c=0; c<MBSIZE; c++)
      {
        for(d=0; d<MBSIZE; d++)
        {
          dataAC[a*MBSIZE+c][b*MBSIZE+d] = ACBuffer[c][d];
        }
      }
      for(c=0; c<4; c++)
      {
        for(d=0; d<4; d++)
        {
          dataDC[a*4+c][b*4+d] = DCBuffer[c][d];
        }
      }
    }
  }
  free_mem2Dint(ACBuffer);
  free_mem2Dint(DCBuffer);
}

/*!
 ***********************************************************************
 * \brief
 *    performs the forward discrete cosine transform in an image
 * \param data
 *    pointer to matrix image data
 * \param hDCTBlocks
 *    number of blocks relative to the height of the image (vertical blocks)
 * \param wDCTBlocks
 *    number of blocks relative to the width of the image (horizontal blocks)
 * \para Notes
 *    the first cycle must be speed up with memcpy
 * \return
 *    none
 ***********************************************************************
 */

void calcftrans_chr(int ** dataAC, int * dataDC,int hDCTBlocks,int wDCTBlocks)
{
  int a,b,c,d;
  int **ACBuffer;
  int DCBuffer[4];
   //temporary buffer for MB store
  get_mem2Dint(&ACBuffer, BSIZE, BSIZE);

  for(a=0; a<hDCTBlocks; a++)
  {
    for(b=0; b<wDCTBlocks; b++)
    {
      for(c=0; c<BSIZE; c++)
      {
        for(d=0; d<BSIZE; d++)
        {
          ACBuffer[c][d] = dataAC[a*BSIZE+c][b*BSIZE+d];
        }
      }
      H26L_transChr(ACBuffer,DCBuffer);
      for(c=0; c<BSIZE; c++)
      {
        for(d=0; d<BSIZE; d++)
        {
          dataAC[a*BSIZE+c][b*BSIZE+d] = ACBuffer[c][d];
        }
      }
      for(c=0; c<4; c++)
      {
        dataDC[a*4*wDCTBlocks+b*4+c] = DCBuffer[c];
      }
    }
  }
  free_mem2Dint(ACBuffer);
}

/*!
 ***********************************************************************
 * \brief
 *    Initialize bitstream buffer
 * \param pBitstream
 *    Bitstream buffer
 * \return
 *    none
 ***********************************************************************
 */
void initializeBitstream(Bitstream *pBitstream)
{
  pBitstream->byte_pos = 0;
  pBitstream->bits_to_go = 8;
  pBitstream->byte_buf = 0;
  pBitstream->write_flag = FALSE;
}

/*!
 ***********************************************************************
 * \brief
 *    writes a int to the pBitstream (of Bitstream type) buffer
 * \param data
 *    int to write to the buffer
 * \param numBits
 *    number of bits to write... the rest of the bits are ignored
 * \param pBitstream
 *    Bitstream buffer
 * \return
 *    none
 ***********************************************************************
 */
void writeIntFGS(unsigned int data, int nbits, Bitstream *pBitstream)
{
  int div,remainder;
  if(nbits > 32)
  {
    printf("WriteIntFGS: numBits is zero or is greater than 32\n");
    exit(1);
  }
  if(nbits == 0) return;
  div = (int)(nbits/8);
  remainder = nbits%8;
  switch(div)  //eliminate bits that don't matter
  {
  case 0:
    writeByteFGS((byte)(0x0FF & data),remainder,pBitstream);
    break;
  case 1:
    writeByteFGS((byte)((0x0FF00 & data) >> 8),remainder,pBitstream);
    writeByteFGS((byte)(0x0FF & data),8,pBitstream);
    break;
  case 2:
    writeByteFGS((byte)((0x0FF0000 & data) >> 16),remainder,pBitstream);
    writeByteFGS((byte)((0x0FF00 & data) >> 8),8,pBitstream);
    writeByteFGS((byte)(0x0FF & data),8,pBitstream);
    break;
  case 3:
    writeByteFGS((byte)((0x0FF000000 & data) >> 24),remainder,pBitstream);
    writeByteFGS((byte)((0x0FF0000 & data) >> 16),8,pBitstream);
    writeByteFGS((byte)((0x0FF00 & data) >> 8),8,pBitstream);
    writeByteFGS((byte)(0x0FF & data),8,pBitstream);
    break;
  case 4:
    writeByteFGS((byte)((0x0FF000000 & data) >> 24),8,pBitstream);
    writeByteFGS((byte)((0x0FF0000 & data) >> 16),8,pBitstream);
    writeByteFGS((byte)((0x0FF00 & data) >> 8),8,pBitstream);
    writeByteFGS((byte)(0x0FF & data),8,pBitstream);
    break;
  }
}

/*!
 ***********************************************************************
 * \brief
 *    writes a byte to the pBitstream (of Bitstream type) buffer
 * \param data
 *    byte to write to the buffer
 * \param numBits
 *    number of bits to write... the rest of the bits are ignored
 * \param pBitstream
 *    Bitstream buffer
 * \para Notes
 *    cycles unfolded.
 * \return
 *    none
 ***********************************************************************
 */
void writeByteFGS(byte data,unsigned int numBits, Bitstream *pBitstream)
{
  int temp;
  if(numBits == 0) return;
  if(numBits > 8)
  {
    printf("writeByteFGS: numBits exceeds 8\n");
    exit(1);
  }
#ifdef FGS_TRACE
/*  if(gtr == 55)
  {
    fprintf(pFGSData->pFGSTraceFile,"Data:%x Lenght:%d\n",data,numBits);
    fflush(pFGSData->pFGSTraceFile);
  }*/
#endif
  data = data & mask_bits[numBits-1];
  temp = pBitstream->bits_to_go - numBits;

  if(temp < 0)      //trying to put more bits that the byte_buf can handle => flush to stream buffer 
  {
    pBitstream->byte_buf = (data >> (numBits - pBitstream->bits_to_go)) | pBitstream->byte_buf;
    pBitstream->streamBuffer[pBitstream->byte_pos] = pBitstream->byte_buf;
    pBitstream->byte_pos++;
    pBitstream->byte_buf = 0;
    pBitstream->bits_to_go = 8 - ABS(temp);
    numBits = ABS(temp);
  }
  else
  {
    pBitstream->bits_to_go = temp;
  }
  if(numBits < 8)
  {
    pBitstream->byte_buf = (data << pBitstream->bits_to_go) | pBitstream->byte_buf;
  }
  else
  {
    pBitstream->streamBuffer[pBitstream->byte_pos] = data;
    pBitstream->byte_pos++;
    pBitstream->byte_buf = 0;
    pBitstream->bits_to_go = 8;
  }
}

/*!
 ***********************************************************************
 * \brief
 *    Flushes the bitstream buffer to disk
 * \param pBitstream
 *    Bitstream buffer
 * \para file
 *    File to flush the buffer
 * \para Notes
 *    Usually is done when a frame is coded (the frame buffer is usually large)
 * \return
 *    none
 ***********************************************************************
 */

void flushBuffer(Bitstream *pBitstream,FILE *file)
{
  insertStuffing(pBitstream);
  if(pBitstream->bits_to_go != 0 && pBitstream->bits_to_go != 8)
  {
    printf("Error in flush buffer\n");
    exit(1);
  }
  if(pBitstream->bits_to_go == 0)
  {
    pBitstream->streamBuffer[pBitstream->byte_pos] = pBitstream->byte_buf;
    pBitstream->byte_pos++;
  }
  pBitstream->byte_buf = 0;
  pBitstream->bits_to_go = 8;
  //flush stream_buffer
  fwrite(pBitstream->streamBuffer,sizeof(byte),pBitstream->byte_pos,file);
  pBitstream->byte_pos = 0;
  pBitstream->write_flag = TRUE;
}

/*!
 ***********************************************************************
 * \brief
 *    Inserts stuffing bytes
 * \param pBitstream
 *    Bitstream buffer
 * \para file
 *    File to flush the buffer
 * \para Notes
 *    for byte aligned writes
 * \return
 *    none
 ***********************************************************************
 *
 *                                             0
 *                                            01
 *                                           011
 *                                          0111
 *                                        0 1111
 *                                       01 1111
 *                                      011 1111
 *                                     0111 1111
 ***********************************************************************
 */

void insertStuffing(Bitstream *pBitstream)
{
  switch(pBitstream->bits_to_go)
  {
  case 0:
    writeByteFGS(0x07F,8,pBitstream);
    break;
  case 1:
    writeByteFGS(0x00,1,pBitstream);
    break;
  case 2:
    writeByteFGS(0x01,2,pBitstream);
    break;
  case 3:
    writeByteFGS(0x03,3,pBitstream);
    break;
  case 4:
    writeByteFGS(0x07,4,pBitstream);
    break;
  case 5:
    writeByteFGS(0x0F,5,pBitstream);
    break;
  case 6:
    writeByteFGS(0x01F,6,pBitstream);
    break;
  case 7:
    writeByteFGS(0x03F,7,pBitstream);
    break;
  case 8:
    writeByteFGS(0x07F,8,pBitstream);
    break;
  }
}


void debug_residue(ImageParameters *pimg)
{
  int a,b;

  for(a=0; a<(pimg->height>>2); a++)
  {
    for(b=0; b<(pimg->width>>2); b++)
    {
      fwrite(&(pFGSData->Ydiff_dc[a][b]),sizeof(int),1,pFGSData->pFGSFile);
    }
  }
  
  for(a=0; a<(pimg->height_cr>>2)*(pimg->width_cr>>2); a++)
  {
    fwrite(&(pFGSData->Udiff_dc[a]),sizeof(int),1,pFGSData->pFGSFile);
  }
  
  for(a=0; a<(pimg->height_cr>>2)*(pimg->width_cr>>2); a++)
  { 
    fwrite(&(pFGSData->Vdiff_dc[a]),sizeof(int),1,pFGSData->pFGSFile);
  }

  for(a=0; a<pimg->height; a++)
  {
      for(b=0; b<pimg->width; b++)
      {
        fwrite(&(pFGSData->Ydiff_ac[a][b]),sizeof(int),1,pFGSData->pFGSFile);
      }
  }
  
  for(a=0; a<pimg->height_cr; a++)
  {
    for(b=0; b<pimg->width_cr; b++)
    {
      fwrite(&(pFGSData->Udiff_ac[a][b]),sizeof(int),1,pFGSData->pFGSFile);
    }
  }
  
  for(a=0; a<pimg->height_cr; a++)
  {
    for(b=0; b<pimg->width_cr; b++)
    {
      fwrite(&(pFGSData->Vdiff_ac[a][b]),sizeof(int),1,pFGSData->pFGSFile);
    }
  }
}